#
# Copyright (C) 2002, 2003, 2004, 2007 Red Hat, Inc.
# Authors:
# Thomas Woerner <twoerner@redhat.com>
# Chris Lumens <clumens@redhat.com>
# Brent Fox <bfox@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import gtk
import gtk.glade
import gobject
import sys
import os
import socket
import types
import gtk_label_autowrap
import etc_services

from fw_config import *
import fw_services
from fw_functions import getPortID, getServiceName, getPortRange
from gtk_cellrenderercheck import *
from gtk_treeviewtooltips import TreeViewTooltips
from fw_parser import *
from fw_sysconfig import *

# get icon theme
icon_theme = gtk.icon_theme_get_default()

# load xml file, initialize icon theme and set LOKKIT_PROG
if os.access(GLADE_NAME, os.F_OK):
    xml = gtk.glade.XML(GLADE_NAME, domain=domain)
    icon_theme.append_search_path("./pixmaps")
    LOKKIT_PROG = "./lokkit"
else:
    xml = gtk.glade.XML(DATADIR + "/" + GLADE_NAME, domain=domain)
    icon_theme.append_search_path(DATADIR)
    LOKKIT_PROG = "/usr/sbin/lokkit"

class ui:
    # You must specify a runPriority for the firstboot module order
    runPriority = 50
    moduleName = _("Firewall")
    commentTag = _("Configure firewall rules")
    shortMessage = _("You can use a firewall to allow access to specific "
                     "services on your computer from other computers and "
                     "prevent unauthorized access from the outside world.  "
                     "Which services, if any, do you wish to allow access to?")

    def destroy(self, args):
        gtk.main_quit()

    def __init__(self):
        self.xml = xml
        self.dirty = False
        self.enabled = True
        self.doDebug = None
        self.custom_file = None
        self.modified = False
        self.clean_config = True
        self.config = None

    def setupScreen(self):
        (width, height) = gtk.icon_size_lookup(gtk.ICON_SIZE_MENU)
        size = min(width, height)
        self.icon = icon_theme.load_icon("%s_48" % APP_NAME, 48, 0)

        self.mainWindow = self.xml.get_widget("mainWindow")
        self.mainWindow.set_icon(self.icon)
        gtk_label_autowrap.set_autowrap(self.mainWindow)

        self.mainHPaned = self.xml.get_widget("mainHPaned")

        self.notebook = self.xml.get_widget("mainNotebook")
        self.mainVBox = self.xml.get_widget("mainVBox")

        self.menu_apply = self.xml.get_widget("menu_apply")
        self.menu_apply.connect("activate", self.apply)
        self.menu_apply.set_sensitive(False)
        self.menu_reload = self.xml.get_widget("menu_reload")
        self.menu_reload.connect("activate", self.readFile)
        self.menu_quit = self.xml.get_widget("menu_quit")
        self.menu_quit.connect("activate", self.quit)

        self.menu_wizard = self.xml.get_widget("menu_wizard")
        self.menu_wizard.connect("activate", self.startWizard)

        self.menu_enable = self.xml.get_widget("menu_enable")
        self.menu_enable.connect("activate", self.firewall_enable)
        self.menu_disable = self.xml.get_widget("menu_disable")
        self.menu_disable.connect("activate", self.firewall_disable)

        self.menu_defaults = self.xml.get_widget("menu_defaults")
        self.menu_defaults.connect("activate", self.defaults)

        self.menu_about = self.xml.get_widget("menu_about")
        self.menu_about.connect("activate", self.about)

        self.wizardToolButton = self.xml.get_widget("wizardToolButton")
        self.wizardToolButton.connect("clicked", self.startWizard)

        self.applyToolButton = self.xml.get_widget("applyToolButton")
        self.applyToolButton.connect("clicked", self.apply)
        self.applyToolButton.set_sensitive(False)
        self.reloadToolButton = self.xml.get_widget("reloadToolButton")
        self.reloadToolButton.connect("clicked", self.readFile)

        self.enableToolButton = self.xml.get_widget("enableToolButton")
        self.enableToolButton.connect("clicked", self.firewall_enable)
        self.disableToolButton = self.xml.get_widget("disableToolButton")
        self.disableToolButton.connect("clicked", self.firewall_disable)

        self.aboutDialog = self.xml.get_widget("aboutDialog")
#        self.aboutDialog.set_version("0.1")
        self.aboutDialog.set_authors(AUTHORS)
        self.aboutDialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        self.aboutDialog.set_transient_for(self.mainWindow)
        self.aboutDialog.set_modal(True)
        self.aboutDialog.set_icon(self.icon)

        self.wizardDialog = self.xml.get_widget("wizardDialog")
        gtk_label_autowrap.set_autowrap(self.wizardDialog)
        self.wizardDialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        self.wizardDialog.set_transient_for(self.mainWindow)
        self.wizardDialog.set_modal(True)
        self.wizardDialog.set_icon(self.icon)

        self.wizardNotebook = self.xml.get_widget("wizardNotebook")

        self.wizardDialogOKButton = self.xml.get_widget("wizardDialogOKButton")
        self.wizardDialogBackButton = self.xml.get_widget( \
            "wizardDialogBackButton")
        self.wizardDialogBackButton.connect("clicked", self.wizard_tab_back)
        self.wizardDialogForwardButton = self.xml.get_widget( \
            "wizardDialogForwardButton")
        self.wizardDialogForwardButton.connect("clicked",
                                               self.wizard_tab_forward)

        self.wizardNetworkComboBox = self.xml.get_widget( \
            "wizardNetworkComboBox")
        self.wizardNetworkComboBox.connect("changed",
                                           self.change_network_combo_cb)
        self.wizardConnectionsComboBox = self.xml.get_widget( \
            "wizardConnectionsComboBox")
        self.wizardSkillComboBox = self.xml.get_widget("wizardSkillComboBox")
        self.wizardCleanupCheckButton = self.xml.get_widget( \
            "wizardCleanupCheckButton")
        
        self.modifiedLabel = self.xml.get_widget("modifiedLabel")

        self.stagesView = self.xml.get_widget("stagesView")
        self.stagesView.get_selection().connect( \
            "changed", self.change_stage_cb)
        self.stagesStore = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_INT,
                                         gobject.TYPE_BOOLEAN)
        self.stagesView.set_model(self.stagesStore)

        col = gtk.TreeViewColumn("", gtk.CellRendererText(), text=0,
                                 sensitive=2)
        self.stagesView.append_column(col)

        for i in xrange(self.notebook.get_n_pages()):
            tab = self.notebook.get_nth_page(i)
            if tab.get_property("visible"):
                self.stagesStore.append([self.notebook.get_tab_label_text(tab),
                                         i, True])
        self.stagesView.get_selection().select_path(0)
        self.notebook.set_show_tabs(False)

        self.otherPortsView = self.xml.get_widget("otherPortsView")
        self.otherPortsView.get_selection().connect( \
            "changed", self.change_ports_selection_cb)
        self.addPortButton = self.xml.get_widget("addPortButton")
        self.removePortButton = self.xml.get_widget("removePortButton")
        self.removePortButton.set_sensitive(False)

        self.statusLabel = self.xml.get_widget("statusLabel")

        # Set up trusted services
        self.trustedView = self.xml.get_widget("trustedView")
        self.trustedView.get_selection().set_mode(gtk.SELECTION_NONE)
        # Columns: trusted, service, port/protocol, inconsistent
        self.trustedStore = gtk.ListStore(gobject.TYPE_BOOLEAN,
                                          gobject.TYPE_STRING,
                                          gobject.TYPE_STRING,
                                          gobject.TYPE_BOOLEAN)
        self.trustedSort = gtk.TreeModelSort(self.trustedStore)
        self.trustedView.set_model(self.trustedSort)
        self.trustedTooltips = TreeViewTooltips(self.trustedView,
                                                self.set_trusted_tooltip)
        self.trustedTooltips.set_property("hide-delay", 60*1000)
        
        colToggle = CellRendererCheck()
#        colToggle = gtk.CellRendererToggle()
#        colToggle.set_property("indicator-size", 14)
        colToggle.connect("toggled", self.view_toggle_cb, self.trustedSort, 0)

        col = gtk.TreeViewColumn('', colToggle, active=0, inconsistent=3)
        self.trustedView.append_column(col)

        col = gtk.TreeViewColumn(_("Service"), gtk.CellRendererText(), text=1)
        col.set_clickable(True)
        col.set_sort_indicator(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.trustedView)
        self.trustedView.append_column(col)
        self.trustedSort.set_sort_column_id(1, gtk.SORT_ASCENDING)

        col = gtk.TreeViewColumn(_("Port/Protocol"), gtk.CellRendererText(),
                                 markup=2)
        col.set_clickable(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.trustedView)
        self.trustedView.append_column(col)

        self.trustedView.set_search_column(1)

        # Set up the view and columns for the Other Ports section.
        self.otherPortsStore = gtk.ListStore(gobject.TYPE_STRING,
                                             gobject.TYPE_STRING,
                                             gobject.TYPE_STRING)
        self.otherPortsViewSort = gtk.TreeModelSort(self.otherPortsStore)
        self.otherPortsView.set_model(self.otherPortsViewSort)

        col = gtk.TreeViewColumn(_("Port"), gtk.CellRendererText(), text=0)
        col.set_sort_indicator(True)
        self.otherPortsViewSort.set_sort_column_id(0, gtk.SORT_ASCENDING)
        col.set_clickable(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.otherPortsView)
        self.otherPortsView.append_column(col)

        col = gtk.TreeViewColumn(_("Protocol"), gtk.CellRendererText(), text=1)
        col.set_clickable(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.otherPortsView)
        self.otherPortsView.append_column(col)

        col = gtk.TreeViewColumn(_("Service"), gtk.CellRendererText(), text=2)
        col.set_clickable(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.otherPortsView)
        self.otherPortsView.append_column(col)
        self.addPortButton.connect("clicked", self.add_port_cb)
        self.removePortButton.connect("clicked", self.remove_port_cb)

        # Add port dialog
        self.addPortDialog = self.xml.get_widget("addPortDialog")
        gtk_label_autowrap.set_autowrap(self.addPortDialog)
        self.addPortDialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        self.addPortDialog.set_transient_for(self.mainWindow)
        self.portDialogOkButton = self.xml.get_widget("portDialogOkButton")
        self.portDialogOkButton.set_sensitive(False)
        self.portEntry = self.xml.get_widget("portEntry")
        self.portEntry.connect("changed", self.port_entry_changed_cb)
        self.portUserCheckButton = self.xml.get_widget("portUserCheckButton")
        self.portUserCheckButton.connect("clicked", self.toggle_port_user_cb)
        self.portUserTable = self.xml.get_widget("portUserTable")

        # Ports from etc services for add port dialog
        self.etcServicesSW = self.xml.get_widget("etcServicesSW")
        self.etcServicesView = self.xml.get_widget("etcServicesView")
        self.etcServicesView.get_selection().connect( \
            "changed", self.etc_services_selection_cb)
        self.etcServicesView.connect("row-activated",
                                     self.etc_services_selection_ok_cb)
        self.etcServicesStore = gtk.ListStore(gobject.TYPE_INT,
                                              gobject.TYPE_STRING,
                                              gobject.TYPE_STRING)
        self.etcServicesSort = gtk.TreeModelSort(self.etcServicesStore)
        self.etcServicesView.set_model(self.etcServicesSort)

        col = gtk.TreeViewColumn(_("Port"), gtk.CellRendererText(), text=0)
        col.set_clickable(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.etcServicesView)
        col.set_sort_indicator(True)
        self.etcServicesSort.set_sort_column_id(0, gtk.SORT_ASCENDING)
        self.etcServicesView.append_column(col)

        col = gtk.TreeViewColumn(_("Protocol"), gtk.CellRendererText(), text=1)
        col.set_clickable(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.etcServicesView)
        self.etcServicesView.append_column(col)

        col = gtk.TreeViewColumn(_("Service"), gtk.CellRendererText(), text=2)
        col.set_clickable(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.etcServicesView)
        self.etcServicesView.append_column(col)

        for port in etc_services.services:
            if port.getId() <= 1024 and port.getProtocol() in [ "tcp", "udp" ]:
                self.etcServicesStore.append([port.getId(), port.getProtocol(),
                                              port.getName()])
        
        # Add options to the protocol combo box.
        self.protoCombo = self.xml.get_widget("protoComboBox")
        self.protoCombo.append_text("tcp")
        self.protoCombo.append_text("udp")
        self.protoCombo.set_active(0)

        # Set up trusted interfaces
        self.interfaceView = self.xml.get_widget("interfaceView")
        self.interfaceView.get_selection().set_mode(gtk.SELECTION_NONE)
        # Columns: active, interface, description, clicked, sensitive
        self.interfaceStore = gtk.TreeStore(\
            gobject.TYPE_BOOLEAN, gobject.TYPE_STRING, gobject.TYPE_STRING,
            gobject.TYPE_BOOLEAN, gobject.TYPE_BOOLEAN)
        self.interfaceSort = gtk.TreeModelSort(self.interfaceStore)
        self.interfaceView.set_model(self.interfaceSort)

        colToggle = CellRendererCheck()
#        colToggle = gtk.CellRendererToggle()
#        colToggle.set_property("indicator-size", 14)
        colToggle.connect ("toggled", self.view_toggle_cb, self.interfaceSort,
                           0, 3, 4)
        colToggle.connect ("toggled", self.view_sensitive_children_cb,
                           self.interfaceSort, 0, 3, 4)
        col = gtk.TreeViewColumn("", colToggle, active=0, sensitive=4)
        self.interfaceView.append_column(col)

        col = gtk.TreeViewColumn(_("Interface"), gtk.CellRendererText(),
                                 text=1, sensitive=4)
        col.set_clickable(True)
        col.set_sort_indicator(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.interfaceView)
        self.interfaceView.append_column(col)
        self.interfaceView.set_expander_column(col)
        self.interfaceSort.set_sort_column_id(1, gtk.SORT_ASCENDING)

        col = gtk.TreeViewColumn(_("Description"), gtk.CellRendererText(),
                                 text=2, sensitive=4)
        self.interfaceView.append_column(col)


        self.masqueradeView = self.xml.get_widget("masqueradeView")
        self.masqueradeView.get_selection().set_mode(gtk.SELECTION_NONE)
        # Columns: active, interface, description, clicked, sensitive
        self.masqueradeStore = gtk.TreeStore(\
            gobject.TYPE_BOOLEAN, gobject.TYPE_STRING, gobject.TYPE_STRING,
            gobject.TYPE_BOOLEAN, gobject.TYPE_BOOLEAN)
        self.masqueradeSort = gtk.TreeModelSort(self.masqueradeStore)
        self.masqueradeView.set_model(self.masqueradeSort)

        colToggle = CellRendererCheck()
#        colToggle = gtk.CellRendererToggle()
#        colToggle.set_property("indicator-size", 14)
        colToggle.connect ("toggled", self.view_toggle_cb, self.masqueradeSort,
                           0, 3, 4)
        colToggle.connect ("toggled", self.view_sensitive_children_cb,
                           self.masqueradeSort, 0, 3, 4)
        col = gtk.TreeViewColumn("", colToggle, active=0, sensitive=4)
        self.masqueradeView.append_column(col)

        col = gtk.TreeViewColumn(_("Interface"), gtk.CellRendererText(),
                                 text=1, sensitive=4)
        col.set_clickable(True)
        col.set_sort_indicator(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.masqueradeView)
        self.masqueradeView.append_column(col)
        self.masqueradeView.set_expander_column(col)
        self.masqueradeSort.set_sort_column_id(1, gtk.SORT_ASCENDING)

        col = gtk.TreeViewColumn(_("Description"), gtk.CellRendererText(),
                                 text=2, sensitive=4)
        self.masqueradeView.append_column(col)

        # custom rules chooser
        self.customRulesChooser = self.xml.get_widget("customRulesChooser")
        self.customRulesChooser.connect("selection-changed",
                                        self.custom_filechooser_cb)

        # custom rules files
        self.customView = self.xml.get_widget("customView")
        # Columns: active, interface, description, clicked, sensitive
        self.customStore = gtk.ListStore(\
            gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_STRING)
        self.customSort = gtk.TreeModelSort(self.customStore)
        self.customView.set_model(self.customSort)

        col = gtk.TreeViewColumn(_("Type"), gtk.CellRendererText(),
                                 text=0)
        col.set_clickable(True)
        col.set_sort_indicator(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.customView)
        self.customView.append_column(col)
        self.customSort.set_sort_column_id(0, gtk.SORT_ASCENDING)

        col = gtk.TreeViewColumn(_("Table"), gtk.CellRendererText(),
                                 text=1)
        col.set_clickable(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.customView)
        self.customView.append_column(col)

        col = gtk.TreeViewColumn(_("Filename"), gtk.CellRendererText(),
                                 text=2)
        col.set_clickable(True)
        col.connect("clicked", self.toggle_view_sort_cb, self.customView)
        self.customView.append_column(col)

        # add/remove custom files
        self.addCustomButton = self.xml.get_widget("addCustomButton")
        self.addCustomButton.connect("clicked", self.add_custom_cb)
        self.removeCustomButton = self.xml.get_widget("removeCustomButton")
        self.removeCustomButton.set_sensitive(False)
        self.removeCustomButton.connect("clicked", self.remove_custom_cb)

        self.customView.get_selection().connect( \
            "changed", self.change_custom_selection_cb)

        self.addCustomDialog = self.xml.get_widget("addCustomDialog")
        gtk_label_autowrap.set_autowrap(self.addCustomDialog)
        self.addCustomDialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        self.addCustomDialog.set_transient_for(self.mainWindow)
        self.addCustomDialog.set_modal(True)
        self.addCustomDialog.set_icon(self.icon)

        self.customDialogOkButton = self.xml.get_widget("customDialogOkButton")

        self.addressTypeCombo = self.xml.get_widget("addressTypeCombo")
        self.firewallTableCombo = self.xml.get_widget("firewallTableCombo")

        # activate the firewall
        self.firewall_enable()
        self.setDirty(False)

    def setDirty(self, val):
        self.dirty = val
        self.menu_apply.set_sensitive(self.dirty)
        self.applyToolButton.set_sensitive(self.dirty)
        if val:
            self.setModified()
            self.modifiedLabel.set_text(_("(modified)"))
        else:
            self.modifiedLabel.set_text("")

    def setModified(self, val=True):
        self.modified = val
        self.menu_defaults.set_sensitive(val)

    def firewall_enable(self, *args):
        self.statusLabel.set_markup("<span color='#008800'>" +\
                                    _("The firewall is enabled.") +\
                                    "</span>")
        self.menu_enable.set_sensitive(0)
        self.menu_disable.set_sensitive(1)
        self.enableToolButton.set_sensitive(0)
        self.disableToolButton.set_sensitive(1)
        self.enabled = True
        self.mainHPaned.set_sensitive(True)
        self.setDirty(True)

    def firewall_disable(self, *args):
        self.statusLabel.set_markup("<span color='#cc0000'>" +\
                                    _("The firewall is disabled.") +\
                                    "</span>")
        self.menu_enable.set_sensitive(1)
        self.menu_disable.set_sensitive(0)
        self.enableToolButton.set_sensitive(1)
        self.disableToolButton.set_sensitive(0)
        self.enabled = False
        self.mainHPaned.set_sensitive(False)
        self.setDirty(True)

    def set_trusted_tooltip(self, model, path, col):
        iter = model.get_iter(path)
        name = model.get(iter, 1)[0]
        svc = fw_services.getByName(name)

        text = "<b>" + name + "</b>"
        if svc.description and len(svc.description) > 0:
            text += "\n" + svc.description
        if svc.default:
            flag = _("enabled")
        else:
            flag = _("disabled")
        text += "\n<small><b>" + _("Default:") + "</b> " + flag + "</small>"
        if svc.modules:
            text += "\n<small><b>" + _("Helper modules:") + "</b> " + \
                ", ".join(svc.modules) + "</small>"
        return text

    def custom_filechooser_cb(self, *args):
        filename = self.customRulesChooser.get_filename()
        self.customDialogOkButton.set_sensitive(filename != None)

    def view_toggle_cb(self, toggle, row, store, col, save=-1, sensitive=-1):
        model = store.get_model()
        iter = store.convert_iter_to_child_iter(None, store.get_iter(row))
        if sensitive < 0 or model.get(iter, sensitive)[0]:
            old_val = model.get(iter, 0)[0]
            model.set(iter, col, not old_val)
            if save >= 0:
                model.set(iter, save, not old_val)
            self.setDirty(True)

    def _setSensitive(self, model, iter, col, clicked_col, sensitive_col, val):
        for i in xrange(model.iter_n_children(iter)):
            iter2 = model.iter_nth_child(iter, i)
            clicked = model.get(iter2, clicked_col)[0]

            # set (in)sensitive
            model.set(iter2, sensitive_col, val)

            if val:
                # set old state
                model.set(iter2, col, clicked)
            else:
                # set new state
                model.set(iter2, col, not val)

            if not clicked or clicked != val:
                self._setSensitive(model, iter2, col, clicked_col,
                                   sensitive_col, val)

    def view_sensitive_children_cb(self, toggle, row, store, col, clicked_col,
                                   sensitive_col):
        model = store.get_model()
        iter = store.convert_iter_to_child_iter(None, store.get_iter(row))
        if sensitive_col >= 0 and not model.get(iter, sensitive_col)[0]:
            return
        self._setSensitive(model, iter, col, clicked_col, sensitive_col,
                           toggle.get_active())

    def toggle_view_sort_cb(self, column, view):
        model = view.get_model()
        (id, order) = model.get_sort_column_id()
        _id = view.get_columns().index(column)
        if _id == id:
            # reverse order
            if order == gtk.SORT_ASCENDING:
                order = gtk.SORT_DESCENDING
            else:
                order = gtk.SORT_ASCENDING
        else:
            order = column.get_sort_order()
            view.get_column(id).set_sort_indicator(False)
        model.set_sort_column_id(_id, order)
        column.set_sort_order(order)
        column.set_sort_indicator(True)

    def change_stage_cb(self, selection):
        (model, iter) = selection.get_selected()
        if iter:
            self.notebook.set_current_page(self.stagesStore.get(iter, 1)[0])
        else:
            i = self.notebook.get_current_page()
            selection.select_path(i)

    def change_ports_selection_cb(self, selection):
        (model, iter) = selection.get_selected()
        if iter:
            self.removePortButton.set_sensitive(True)
        else:
            self.removePortButton.set_sensitive(False)

    def change_custom_selection_cb(self, selection):
        (model, iter) = selection.get_selected()
        if iter:
            self.removeCustomButton.set_sensitive(True)
        else:
            self.removeCustomButton.set_sensitive(False)

    def add_port_cb(self, button, *args):
        self.addPortDialog.show_all()
        self.etcServicesView.get_selection().unselect_all()
        self.portEntry.set_text("")
        self.protoCombo.set_active(0)
        self.portUserCheckButton.set_active(False)

        # Loop until we get a valid port string.
        while True:
            result = self.addPortDialog.run()

            if result == gtk.RESPONSE_OK:
                if not self.portUserCheckButton.get_active():
                    selection = self.etcServicesView.get_selection()
                    (model, iter) = selection.get_selected()
                    if not iter:
                        continue
                    iter = model.convert_iter_to_child_iter(None, iter)
                    if not iter:
                        continue
                    port = self.etcServicesStore.get(iter, 0)[0]
                    proto = self.etcServicesStore.get(iter, 1)[0]
                else:
                    port = self.portEntry.get_text()
                    if port == "":
                        continue
                    proto = self.protoCombo.get_active_text()

                ports = getPortRange(port)
                if not (isinstance(ports, types.ListType) or \
                        isinstance(ports, types.TupleType)):
                    self.invalidPortDialog(port)
                    continue
                self._portsStoreAppend(ports, proto)
                self.setDirty(True)
            else:
                self.etcServicesView.get_selection().unselect_all()

            break

        self.addPortDialog.hide()

    def _portsStoreAppend(self, ports, proto):
        if len(ports) == 2:
            name1 = getServiceName(ports[0], proto)
            name2 = getServiceName(ports[1], proto)
            self.otherPortsStore.append(["%d-%d" % (ports[0], ports[1]), proto,
                                         "%s - %s" % (name1, name2)])
        else:
            name = getServiceName(ports[0], proto)
            self.otherPortsStore.append([ports[0], proto, name])

    def remove_port_cb(self, button, *args):
        selection = self.otherPortsView.get_selection()
        (model, sort_iter) = selection.get_selected()

        if sort_iter is None:
            return

        # Convert the iter on the TreeModelSort to an iter on the underlying
        # ListStore so we can delete the right thing.
        iter = model.convert_iter_to_child_iter(None, sort_iter)

        if iter is not None:
            # Need to get past the TreeModelSort to whatever's underneath.
            model.get_model().remove(iter)
            self.setDirty(True)

    def add_custom_cb(self, button, *args):
        self.addCustomDialog.show_all()
        self.addressTypeCombo.set_active(0)
        self.firewallTableCombo.set_active(0)
        self.customRulesChooser.set_filename("")
        self.customDialogOkButton.set_sensitive(False)
        result = self.addCustomDialog.run()
        self.addCustomDialog.hide()
        
        if result == gtk.RESPONSE_OK:
            type = self.addressTypeCombo.get_active_text()
            table = self.firewallTableCombo.get_active_text()
            filename = self.customRulesChooser.get_filename()

            self.setDirty(True)
            self.customStore.append([ type, table, filename ])
 
    def remove_custom_cb(self, button, *args):
        selection = self.customView.get_selection()
        (model, sort_iter) = selection.get_selected()

        if sort_iter is None:
            return

        # Convert the iter on the TreeModelSort to an iter on the underlying
        # ListStore so we can delete the right thing.
        iter = model.convert_iter_to_child_iter(None, sort_iter)

        if iter is not None:
            # Need to get past the TreeModelSort to whatever's underneath.
            model.get_model().remove(iter)
            self.setDirty(True)

    def toggle_port_user_cb(self, check):
        val = check.get_active()
        self.etcServicesSW.set_sensitive(not val)
        self.portUserTable.set_sensitive(val)

    def port_entry_changed_cb(self, entry):
        text = entry.get_text()
        if text:
            self.portDialogOkButton.set_sensitive(True)
        else:
            self.portDialogOkButton.set_sensitive(False)

    def etc_services_selection_ok_cb(self, *args):
        self.portDialogOkButton.clicked()

    def etc_services_selection_cb(self, selection):
        (model, iter) = selection.get_selected()
        if iter:
            self.portDialogOkButton.set_sensitive(True)
        else:
            self.portDialogOkButton.set_sensitive(False)

    def quit(self, *args):
        if self.dirty:
            dialog = gtk.MessageDialog(None, 0, gtk.MESSAGE_WARNING,
                                       gtk.BUTTONS_YES_NO,
                                       _("There are unapplied changes, do you "
                                         "really want to quit?"))
            gtk_label_autowrap.set_autowrap(dialog)
            dialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
            dialog.set_transient_for(self.mainWindow)
            dialog.set_modal(True)
            dialog.set_icon(self.icon)
            dialog.show_all()
            result = dialog.run()
            dialog.destroy()
            if result != gtk.RESPONSE_YES:
                return True
        self.destroy(args)

    def about(self, *args):
        self.aboutDialog.show_all()
        self.aboutDialog.run()
        self.aboutDialog.hide()

    def defaults(self, *args):
        modified = self.modified

        self.loadDefaults()

        if modified:
            self.setDirty(True)
        self.modified = False
        self.menu_defaults.set_sensitive(False)

    def _findPortInStore(self, port_id, proto):
        str = "%s/%s" % (port_id, proto)
        iter = self.otherPortsStore.get_iter_first()
        while iter:
            if port_id == self.otherPortsStore.get_value(iter, 0) and \
                   proto == self.otherPortsStore.get_value(iter, 1):
                return iter
            else:
                iter = self.otherPortsStore.iter_next(iter)

        return None

    def _setInterfaces(self, store, iter, list):
        while iter:
            dev = store.get_value(iter, 1)
            if dev in list:
                list.remove(dev)
                store.set_value(iter, 0, True)
                store.set_value(iter, 3, True)
                self.setModified()
            iter2 = store.iter_nth_child(iter, 0)
            if iter2:
                self._setInterfaces(store, iter2, list)
            iter = store.iter_next(iter)

    def _addDevice(self, store, device, description, enabled):
        iter = None
        _iter = store.get_iter_first()
        while _iter:
            devplus = store.get_value(_iter, 1)
            if devplus[-1] != "+":
                continue
            dev = devplus[:-1]
            if device.startswith(dev):
                iter = _iter
                break
            _iter = store.iter_next(_iter)
        store.append(iter, [enabled, device, description, enabled, True])

    def loadConfig(self, config):
        self.config = config

        if not config:
            return

        self.clean_config = False

        # enabled / disabled
        if not config.enabled:
            self.firewall_disable()
            self.setModified()

        (services, other_ports) = fw_services.getServicesAndPorts(config)

        # trusted services
        for svc in fw_services.service_list:
            enabled = False
            if svc.default:
                if not (hasattr(config, "no_%s" % svc.key) and \
                        getattr(config, "no_%s" % svc.key)):
                    enabled = True
            else:
                if svc.key in services:
                    enabled = True
            store = self.trustedStore
            iter = store.get_iter_first()
            while iter:
                if store.get_value(iter, 1) == svc.name:
                    store.set_value(iter, 0, enabled)
                    self.setModified()
                    break
                iter = store.iter_next(iter)

        # other ports
        for (ports, proto) in other_ports:
            self._portsStoreAppend(ports, proto)
            self.setModified()

        # trusted devices
        devices = [ ]
        if config.trust:
            devices.extend(config.trust)
        # configure trusted interfaces
        self._setInterfaces(self.interfaceStore,
                            self.interfaceStore.get_iter_first(), devices)
        # add remaining devices
        for device in devices:
            self._addDevice(self.interfaceStore, device,
                            _("Not configured"), True)
            self.setModified()
        self.interfaceView.expand_all()

        # masquerading
        masq = [ ]
        if config.masq:
            masq.extend(config.masq)
        # configure masquerading interfaces
        self._setInterfaces(self.masqueradeStore,
                            self.masqueradeStore.get_iter_first(), masq)
        # add remaining devices
        for device in masq:
            self._addDevice(self.masqueradeStore, device,
                            _("Not configured"), True)
            self.setModified()
        self.masqueradeView.expand_all()

        # custom rules
        if config.custom_rules:
            for entry in config.custom_rules:
                self.customStore.append(entry)

        return

    def readFile(self, *args):
        self.loadDefaults()

        config = read_sysconfig_config()
        if not config:
            # no configuration: defaults are not applied, yet
            self.setDirty(True)
            return
        self.loadConfig(config)

        # Check if firewall config files exist
        if config.enabled and not \
                (os.path.exists(IP4TABLES_RULES) and \
                 os.path.isfile(IP4TABLES_RULES) and \
                 os.path.exists(IP6TABLES_RULES) and \
                 os.path.isfile(IP6TABLES_RULES)):
            dialog = gtk.MessageDialog(None, 0, gtk.MESSAGE_WARNING,
                                       gtk.BUTTONS_OK,
                                       _("The firewall configuration "
                                         "is not consistent."))
            text = _("The following files are missing or unusable:") + "\n"
            if not os.path.exists(IP4TABLES_RULES) or  \
                    not os.path.isfile(IP4TABLES_RULES):
                text += "\t%s\n" % IP4TABLES_RULES
            if not os.path.exists(IP6TABLES_RULES) or  \
                    not os.path.isfile(IP6TABLES_RULES):
                text += "\t%s\n" % IP6TABLES_RULES
            text += "\n" + _("Apply your firewall configuration now "
                             "to correct this problem.")
            dialog.format_secondary_text(text)
            gtk_label_autowrap.set_autowrap(dialog)
            dialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
            dialog.set_transient_for(self.mainWindow)
            dialog.set_modal(True)
            dialog.set_icon(self.icon)
            dialog.show_all()
            result = dialog.run()
            dialog.destroy()
            self.setDirty(True)
        else:
            self.setDirty(False)

    def invalidPortDialog(self, port):
        dialog = gtk.MessageDialog(None, 0, gtk.MESSAGE_WARNING,
                                   gtk.BUTTONS_OK)
        dialog.set_markup("<b>" + \
                          _("Port or port range '%s' is not valid.") % port + \
                          "</b>")
        dialog.format_secondary_text(_("Please enter a valid port number, "
                                       "service name or range."))
        gtk_label_autowrap.set_autowrap(dialog)
        dialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        dialog.set_transient_for(self.addPortDialog)
        dialog.set_modal(True)
        dialog.set_icon(self.icon)
        dialog.show_all()
        dialog.run()
        dialog.destroy()

    def confirmDialog(self):
        dialog = gtk.MessageDialog(None, 0, gtk.MESSAGE_WARNING,
                                   gtk.BUTTONS_YES_NO,
                                   _("Clicking the 'Yes' button will override "
                                     "any existing firewall configuration. "
                                     "Are you sure that you want to do this?"))
        gtk_label_autowrap.set_autowrap(dialog)
        dialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        dialog.set_transient_for(self.mainWindow)
        dialog.set_modal(True)
        dialog.set_icon(self.icon)
        dialog.show_all()
        result = dialog.run()
        dialog.destroy()
        return result

    def _getInterfaces(self, store, iter, list):
        while iter:
            dev = store.get_value(iter, 1)
            if store.get_value(iter, 0):
                list.append(dev)
            else:
                iter2 = store.iter_nth_child(iter, 0)
                if iter2:
                    self._getInterfaces(store, iter2, list)
            iter = store.iter_next(iter)

    def startWizard(self, *args):
        self.wizardNotebook.set_show_tabs(False)
        self.wizardNotebook.set_current_page(0)
        self.wizardNetworkComboBox.set_active(1)
        self.wizardConnectionsComboBox.set_active(0)
        self.wizardCleanupCheckButton.set_active(True)
        self.wizardSkillComboBox.set_active(0)
        self.adjust_wizard_buttons()
        self.wizardDialog.show_all()
        while True:
            result = self.wizardDialog.run()
            if result == gtk.RESPONSE_OK:
                # network active?
                if self.wizardNetworkComboBox.get_active() == 0:
                    # no network
                    self.firewall_disable()
                    break
                else:
                    self.firewall_enable()
                # cleanup
                if self.wizardCleanupCheckButton.get_active():
                    self.loadDefaults()
                    self.setDirty(True)
                remove = [ ]
                # connections (important: after load defaults)
                if self.wizardConnectionsComboBox.get_active() == 0:
                    remove.append(_("Trusted Interfaces"))
                    self.interfaceStore.clear()
                    remove.append(_("Masquerading"))
                    self.masqueradeStore.clear()
                # skill level (important: after load defaults)
                if self.wizardSkillComboBox.get_active() == 0:
                    remove.append(_("Custom Rules"))
                    self.customStore.clear()
                # remove items
                for i in xrange(self.notebook.get_n_pages()):
                    page = self.notebook.get_nth_page(i)
                    name = self.notebook.get_tab_label_text(page)
                    if name not in remove:
                        continue
                    page.set_sensitive(False)
                    iter = self.stagesStore.get_iter_first()
                    while iter:
                        if name == self.stagesStore.get_value(iter, 0):
                            self.stagesStore.set_value(iter, 2, False)
                        iter = self.stagesStore.iter_next(iter)

                self.stagesView.get_selection().select_path(0)
                

                break
            elif result == gtk.RESPONSE_CANCEL:
                break
        self.wizardDialog.hide()

    def adjust_wizard_buttons(self):
        i = self.wizardNotebook.get_current_page()
        if i >= 1:
            self.wizardDialogBackButton.set_sensitive(True)
        else:
            self.wizardDialogBackButton.set_sensitive(False)
        if i == self.wizardNotebook.get_n_pages() - 1:
            self.wizardDialogOKButton.set_sensitive(True)
            self.wizardDialogForwardButton.set_sensitive(False)
        elif i < self.wizardNotebook.get_n_pages() - 1:
            self.wizardDialogForwardButton.set_property("visible", True)
            self.wizardDialogForwardButton.set_sensitive(True)
            self.wizardDialogOKButton.set_sensitive(False)
        else:
            self.wizardDialogForwardButton.set_sensitive(False)
        page = self.wizardNotebook.get_nth_page(i)
        name = self.wizardNotebook.get_tab_label_text(page)
        if name == "network" and self.wizardNetworkComboBox.get_active() == 0:
            self.wizardDialogOKButton.set_sensitive(True)
            self.wizardDialogForwardButton.set_sensitive(False)

    def wizard_tab_back(self, *args):
        i = self.wizardNotebook.get_current_page() - 1
        self.wizardNotebook.set_current_page(i)
        self.adjust_wizard_buttons()

    def wizard_tab_forward(self, *args):
        i = self.wizardNotebook.get_current_page() + 1
        self.wizardNotebook.set_current_page(i)
        self.adjust_wizard_buttons()

    def change_network_combo_cb(self, *args):
        i = self.wizardNotebook.get_current_page()
        child = self.wizardNotebook.get_nth_page(i)
        name = self.wizardNotebook.get_tab_label_text(child)
        if name == "network" and self.wizardNetworkComboBox.get_active() == 0:
            self.wizardDialogOKButton.set_sensitive(True)
            self.wizardDialogForwardButton.set_sensitive(False)
        else:
            self.wizardDialogOKButton.set_sensitive(False)
            self.wizardDialogForwardButton.set_sensitive(True)

    def genArgs(self):
        # With the new enabled/disabled behavior, we have to ignore the config
        # file or else you can only ever turn on services.
        args = [ LOKKIT_PROG, '--quiet', '-f' ]

        if not self.dirty:
            return 0

        if self.enabled:
            args.append('--enabled')
        else:
            args.append('--disabled')

        trust = [ ]
        iter = self.interfaceStore.get_iter_first()
        self._getInterfaces(self.interfaceStore, iter, trust)
        for dev in trust:
            args.append("--trust=%s" % dev)

        masquerade = [ ]
        iter = self.masqueradeStore.get_iter_first()
        self._getInterfaces(self.masqueradeStore, iter, masquerade)
        for dev in masquerade:
            args.append("--masq=%s" % dev)

        iter = self.trustedStore.get_iter_first()
        while iter:
            svc = fw_services.getByName( \
                self.trustedStore.get_value(iter, 1))
            if self.trustedStore.get_value(iter, 0):
                if not svc.default:
                    for (port, proto) in svc.ports:
                        args.append("--port=%s:%s" % (port, proto))
                    for module in svc.modules:
                        args.append("--addmodule=%s" % module)
            else:
                if svc.default:
                    args.append("--no-%s" % svc.key)
                for module in svc.modules:
                    args.append("--removemodule=%s" % module)

            iter = self.trustedStore.iter_next(iter)

        model = self.otherPortsView.get_model()
        iter = model.get_iter_first()
        while iter:
            port = model.get_value(iter, 0)
            proto = model.get_value(iter, 1)
            args.append("--port=%s:%s" % (port, proto))
            iter = model.iter_next(iter)

        model = self.customView.get_model()
        iter = model.get_iter_first()
        while iter:
            type = model.get_value(iter, 0)
            table = model.get_value(iter, 1)
            filename = model.get_value(iter, 2)
            args.append("--custom-rules=%s:%s:%s" % (type, table, filename))
            iter = model.iter_next(iter)

        return args

    def apply(self, *args):
        args = self.genArgs()

        if self.confirmDialog() == gtk.RESPONSE_NO:
            return None

        if self.doDebug:
            print "don't call lokkit if in debug mode"
            print " ".join(args)
        else:
            (rfd, wfd) = os.pipe()
            pid = os.fork()
            if pid == 0:
                try:
                    os.close(rfd)
                    fd = os.open("/dev/null", os.O_RDONLY)
                    if fd != 0:
                        os.dup2(fd, 0)
                        os.close(fd)
                    if wfd != 1:
                        os.dup2(wfd, 1)
                        os.close(wfd)
                    os.dup2(1, 2)
                    os.execv(args[0], args)
                finally:
                    os._exit(255)

            os.close(wfd)
            # no need to read in chunks if we don't pass on data to some
            # output func
            cret = ""
            cout = os.read(rfd, 8192)
            while cout:
                cret += cout
                cout = os.read(rfd, 8192)
            os.close(rfd)
            (cpid, status) = os.waitpid(pid, 0)

            # failed to configure firewall, show error message
            if status != 0:
                dialog = gtk.MessageDialog(None, 0, gtk.MESSAGE_ERROR,
                                        gtk.BUTTONS_CLOSE)
                dialog.set_markup("<b>" + _("Configuration failed") + \
                                  "</b>")
                dialog.format_secondary_text(" ".join(args) + "\n\n" + \
                                             cret.strip())
                gtk_label_autowrap.set_autowrap(dialog)
                dialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
                dialog.set_transient_for(self.mainWindow)
                dialog.set_modal(True)
                dialog.set_icon(self.icon)
                dialog.show_all()
                dialog.run()
                dialog.destroy()
                return 1

        self.loadDefaults()
        self.readFile()

        return 0

    def loadDefaults(self):
        self.clean_config = True
        self.firewall_enable()
        
        # make all tabs and the matching labels sensitive
        for i in xrange(self.notebook.get_n_pages()):
            page = self.notebook.get_nth_page(i)
            name = self.notebook.get_tab_label_text(page)
            page.set_sensitive(True)
            iter = self.stagesStore.get_iter_first()
            while iter:
                if name == self.stagesStore.get_value(iter, 0):
                    self.stagesStore.set_value(iter, 2, True)
                iter = self.stagesStore.iter_next(iter)

        # trusted services
        self.trustedStore.clear()
        for svc in fw_services.service_list:
            ports = [ ]
            for port in svc.ports:
                if port[0] != "":
                    port_id = getPortID(port[0])
                else:
                    port_id = ""
                ports.append("%s/%s" % (port_id, port[1]))

            self.trustedStore.append([svc.default, svc.name,
                                      ", ".join(ports), False])

        # other ports
        self.otherPortsStore.clear()

        # interfaces
        self.interfaceStore.clear()
        self.masqueradeStore.clear()
        # global devices
        for dev in STD_DEVICES:
            devplus = "%s+" % dev
            desc = "-- " + _("All %s devices") % dev + " --"

            self._addDevice(self.interfaceStore, devplus, desc, False)
            self._addDevice(self.masqueradeStore, devplus, desc, False)

        devices = { }
        devlist = NCDeviceList.getDeviceList()
        for device in devlist:
            if device.Alias and dev.Alias != "":
                # ignore device aliases: not usable for iptables
                continue
            devices.setdefault(device.Device, [ ]).append(device)

        # add local devices
        for dev in devices:
            desc = ""
            for device in devices[dev]:
                if desc != "":
                    desc += "\n"
                if device.DeviceId != "dhcp" and device.DeviceId != dev:
                    desc += "%s (%s)" % (device.Type, device.DeviceId)
                else:
                    desc += device.Type
                if device.BootProto != "none":
                    desc += ", %s" % device.BootProto
                else:
                    desc += ", %s/%s" % (device.IP, device.Netmask)
                if device.HardwareAddress:
                    desc += ", HWADDR %s" % device.HardwareAddress
                if device.OnBoot:
                    desc += ", ONBOOT"
            self._addDevice(self.interfaceStore, dev, desc, False)
            self._addDevice(self.masqueradeStore, dev, desc, False)
        self.interfaceView.expand_all()
        self.masqueradeView.expand_all()

        # custom file
        self.customStore.clear()

        self.setModified(False)

    def showStartupDialog(self):
        dialog = self.xml.get_widget("startupDialog")
        gtk_label_autowrap.set_autowrap(dialog)
        dialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        dialog.set_transient_for(self.mainWindow)
        dialog.set_modal(True)
        dialog.set_icon(self.icon)
        dialog.show_all()
        # show for 30 seconds
        timer = gobject.timeout_add(30*1000, self.hideStartupDialog, dialog)
        dialog.run()
        dialog.destroy()

    def hideStartupDialog(self, dialog):
        dialog.hide()
        dialog.destroy()

    def launch(self, doDebug = None):
        self.doDebug = doDebug
        self.setupScreen()

        messageLabel = gtk.Label(_(self.shortMessage))
        messageLabel.set_line_wrap(True)
        messageLabel.set_size_request(500, -1)
        messageLabel.set_alignment(0.0, 0.5)

        self.readFile()
        vbox = gtk.VBox(spacing=10)
        vbox.pack_start(messageLabel, expand=False)
        self.mainVBox.reparent(vbox)

        icon = gtk.Image()
        icon.set_from_pixbuf(self.icon)
        return vbox, icon, self.moduleName

    def run(self):
        self.setupScreen()
        self.mainWindow.connect("delete_event", self.quit)
        self.mainWindow.show()
        self.readFile()
#        if self.clean_config:
        self.showStartupDialog()
        gtk.main()
