#
# Copyright (C) 2007 Red Hat, Inc.
# Authors:
# Thomas Woerner <twoerner@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import gtk
import gobject

class CellRendererCheck(gtk.GenericCellRenderer):

    __gproperties__ = {
        "active"      : (gobject.TYPE_BOOLEAN,
                         "Active", "Active", False,
                         gobject.PARAM_READWRITE),
        "inconsistent": (gobject.TYPE_BOOLEAN,
                         "Inconsistent", "Inconsistent", False,
                         gobject.PARAM_READWRITE),
        }

    __gsignals__ = {
        "toggled": (gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE,
                    (gobject.TYPE_STRING, ))
        }

    def __init__(self, icon_active="check-active_16",
                 icon_inactive="check-inactive_16",
                 icon_inconsistent="check-inconsistent_16",
                 icon_size=gtk.ICON_SIZE_MENU):
        self.__gobject_init__()
        self.set_property("mode", gtk.CELL_RENDERER_MODE_ACTIVATABLE)
        self.set_property("xpad", 2)
        self.set_property("ypad", 2)
        self.active = False
        self.inconsistent = False
        self.icon_size = icon_size

        (width, height) = gtk.icon_size_lookup(self.icon_size)
        size = min(width, height)

        icon_theme = gtk.icon_theme_get_default()

        self.pixbuf = { }
        self.pixbuf["active"] = icon_theme.load_icon(icon_active, size, 0)
        self.pixbuf["inactive"] = icon_theme.load_icon(icon_inactive, size, 0)
        self.pixbuf["inconsistent"] = icon_theme.load_icon(icon_inconsistent,
                                                           size, 0)

    def do_get_property(self, pspec):
        return getattr(self, pspec.name)

    def do_set_property(self, pspec, value):
        setattr(self, pspec.name, value)

    def get_active(self):
        return self.active

    def __getPixbuf(self):
        if self.inconsistent:
            return self.pixbuf["inconsistent"]
        else:
            if self.active:
                return self.pixbuf["active"]
            else:
                return self.pixbuf["inactive"]

    def on_activate(self, event, widget, path, bg_area, cell_area, flags):
        self.emit("toggled", path)

    def on_render(self, window, widget, bg_area, cell_area, expose_area, flags):
        (x, y, width, height) = self.on_get_size(widget, cell_area)
        x += cell_area.x + self.get_property("xpad")
        y += cell_area.y + self.get_property("ypad")
        width -= 2 * self.get_property("xpad")
        height -= 2 * self.get_property("ypad")

        rect = cell_area.intersect(gtk.gdk.Rectangle(x, y, width, height))
        rect = expose_area.intersect(rect)

        pix = self.__getPixbuf()
        if not pix:
            return

        if self.get_property("sensitive") == False:
            # make icon insensitive
            pix = gtk.IconSet(pix).render_icon(widget.get_style(),
                                               gtk.TEXT_DIR_NONE,
                                               gtk.STATE_INSENSITIVE,
                                               self.icon_size, None, None)

        window.draw_pixbuf(widget.style.black_gc, pix, rect.x-x, rect.y-y,
                           rect.x, rect.y, rect.width, rect.height,
                           gtk.gdk.RGB_DITHER_NONE, 0, 0)

    def on_get_size(self, widget, cell_area):
        pixbuf = self.__getPixbuf()
        if not pixbuf:
            return (0, 0, 0, 0)

        pixbuf_width  = pixbuf.get_width()
        pixbuf_height = pixbuf.get_height()
        width  = pixbuf_width + self.get_property("xpad") * 2
        height = pixbuf_height + self.get_property("ypad") * 2

        if not cell_area or pixbuf_width < 1 or pixbuf_height < 1:
            return (0, 0, width, height)

        x = self.get_property("xalign") * (cell_area.width - width -
                                           self.get_property("xpad"))
        y = self.get_property("yalign") * (cell_area.height - height -
                                           self.get_property("ypad"))

        return (int(x), int(y), width, height)

gobject.type_register(CellRendererCheck)
